"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.addProfilingExtensionMethods = exports.__PRIVATE__wrapStartTransactionWithProfiling = exports.stopTransactionProfile = exports.maybeProfileTransaction = exports.MAX_PROFILE_DURATION_MS = void 0;
const utils_1 = require("@sentry/utils");
const env_1 = require("./env");
const cpu_profiler_1 = require("./cpu_profiler");
const utils_2 = require("./utils");
const core_1 = require("@sentry/core");
exports.MAX_PROFILE_DURATION_MS = 30 * 1000;
// Takes a transaction and determines if it should be profiled or not. If it should be profiled, it returns the
// profile_id, otherwise returns undefined. Takes care of setting profile context on transaction as well
function maybeProfileTransaction(client, transaction, customSamplingContext) {
    // profilesSampleRate is multiplied with tracesSampleRate to get the final sampling rate. We dont perform
    // the actual multiplication to get the final rate, but we discard the profile if the transaction was sampled,
    // so anything after this block from here is based on the transaction sampling.
    if (!transaction.sampled) {
        return;
    }
    // Client and options are required for profiling
    if (!client) {
        if ((0, env_1.isDebugBuild)()) {
            utils_1.logger.log('[Profiling] Profiling disabled, no client found.');
        }
        return;
    }
    const options = client.getOptions();
    if (!options) {
        if ((0, env_1.isDebugBuild)()) {
            utils_1.logger.log('[Profiling] Profiling disabled, no options found.');
        }
        return;
    }
    const profilesSampler = options.profilesSampler;
    let profilesSampleRate = options.profilesSampleRate;
    // Prefer sampler to sample rate if both are provided.
    if (typeof profilesSampler === 'function') {
        profilesSampleRate = profilesSampler(Object.assign({ transactionContext: transaction.toContext() }, customSamplingContext));
    }
    // Since this is coming from the user (or from a function provided by the user), who knows what we might get. (The
    // only valid values are booleans or numbers between 0 and 1.)
    if (!(0, utils_2.isValidSampleRate)(profilesSampleRate)) {
        if ((0, env_1.isDebugBuild)()) {
            utils_1.logger.warn('[Profiling] Discarding profile because of invalid sample rate.');
        }
        return;
    }
    // if the function returned 0 (or false), or if `profileSampleRate` is 0, it's a sign the profile should be dropped
    if (!profilesSampleRate) {
        if ((0, env_1.isDebugBuild)()) {
            utils_1.logger.log(`[Profiling] Discarding profile because ${typeof profilesSampler === 'function'
                ? 'profileSampler returned 0 or false'
                : 'a negative sampling decision was inherited or profileSampleRate is set to 0'}`);
        }
        return;
    }
    // Now we roll the dice. Math.random is inclusive of 0, but not of 1, so strict < is safe here. In case sampleRate is
    // a boolean, the < comparison will cause it to be automatically cast to 1 if it's true and 0 if it's false.
    const sampled = Math.random() < profilesSampleRate;
    // Check if we should sample this profile
    if (!sampled) {
        if ((0, env_1.isDebugBuild)()) {
            utils_1.logger.log(`[Profiling] Discarding profile because it's not included in the random sample (sampling rate = ${Number(profilesSampleRate)})`);
        }
        return;
    }
    const profile_id = (0, utils_1.uuid4)();
    cpu_profiler_1.CpuProfilerBindings.startProfiling(profile_id);
    if ((0, env_1.isDebugBuild)()) {
        utils_1.logger.log('[Profiling] started profiling transaction: ' + transaction.name);
    }
    // set transaction context - do this regardless if profiling fails down the line
    // so that we can still see the profile_id in the transaction context
    return profile_id;
}
exports.maybeProfileTransaction = maybeProfileTransaction;
/**
 * Stops the profiler for profile_id and returns the profile
 * @param transaction
 * @param profile_id
 * @returns
 */
function stopTransactionProfile(transaction, profile_id) {
    // Should not happen, but satisfy the type checker and be safe regardless.
    if (!profile_id) {
        return null;
    }
    const profile = cpu_profiler_1.CpuProfilerBindings.stopProfiling(profile_id);
    if ((0, env_1.isDebugBuild)()) {
        utils_1.logger.log('[Profiling] stopped profiling of transaction: ' + transaction.name);
    }
    // In case of an overlapping transaction, stopProfiling may return null and silently ignore the overlapping profile.
    if (!profile) {
        if ((0, env_1.isDebugBuild)()) {
            utils_1.logger.log('[Profiling] profiler returned null profile for: ' + transaction.name, 'this may indicate an overlapping transaction or a call to stopProfiling with a profile title that was never started');
        }
        return null;
    }
    // Assign profile_id to the profile
    profile.profile_id = profile_id;
    return profile;
}
exports.stopTransactionProfile = stopTransactionProfile;
// Wraps startTransaction and stopTransaction with profiling related logic.
// startProfiling is called after the call to startTransaction in order to avoid our own code from
// being profiled. Because of that same reason, stopProfiling is called before the call to stopTransaction.
function __PRIVATE__wrapStartTransactionWithProfiling(startTransaction) {
    return function wrappedStartTransaction(transactionContext, customSamplingContext) {
        const transaction = startTransaction.call(this, transactionContext, customSamplingContext);
        // Client is required if we want to profile
        const client = this.getClient();
        if (!client) {
            return transaction;
        }
        // Check if we should profile this transaction. If a profile_id is returned, then profiling has been started.
        const profile_id = maybeProfileTransaction(client, transaction, customSamplingContext);
        if (!profile_id) {
            return transaction;
        }
        // A couple of important things to note here:
        // `CpuProfilerBindings.stopProfiling` will be scheduled to run in 30seconds in order to exceed max profile duration.
        // Whichever of the two (transaction.finish/timeout) is first to run, the profiling will be stopped and the gathered profile
        // will be processed when the original transaction is finished. Since onProfileHandler can be invoked multiple times in the
        // event of an error or user mistake (calling transaction.finish multiple times), it is important that the behavior of onProfileHandler
        // is idempotent as we do not want any timings or profiles to be overriden by the last call to onProfileHandler.
        // After the original finish method is called, the event will be reported through the integration and delegated to transport.
        let profile = null;
        const options = client.getOptions();
        // Not intended for external use, hence missing types, but we want to profile a couple of things at Sentry that
        // currently exceed the default timeout set by the SDKs.
        const maxProfileDurationMs = 
        // @ts-expect-error maxProfileDurationMs is not intended for external use
        (options._experiments && options._experiments.maxProfileDurationMs) || exports.MAX_PROFILE_DURATION_MS;
        // Enqueue a timeout to prevent profiles from running over max duration.
        let maxDurationTimeoutID = global.setTimeout(() => {
            if ((0, env_1.isDebugBuild)()) {
                utils_1.logger.log('[Profiling] max profile duration elapsed, stopping profiling for:', transaction.name);
            }
            profile = stopTransactionProfile(transaction, profile_id);
        }, maxProfileDurationMs);
        // We need to reference the original finish call to avoid creating an infinite loop
        const originalFinish = transaction.finish.bind(transaction);
        // Wrap the transaction finish method to stop profiling and set the profile on the transaction.
        function profilingWrappedTransactionFinish() {
            if (!profile_id) {
                return originalFinish();
            }
            // We stop the handler first to ensure that the timeout is cleared and the profile is stopped.
            if (maxDurationTimeoutID) {
                global.clearTimeout(maxDurationTimeoutID);
                maxDurationTimeoutID = undefined;
            }
            // onProfileHandler should always return the same profile even if this is called multiple times.
            // Always call onProfileHandler to ensure stopProfiling is called and the timeout is cleared.
            if (!profile) {
                profile = stopTransactionProfile(transaction, profile_id);
            }
            // @ts-expect-error profile is not a part of sdk metadata so we expect error until it becomes part of the official SDK.
            transaction.setMetadata({ profile });
            return originalFinish();
        }
        transaction.finish = profilingWrappedTransactionFinish;
        return transaction;
    };
}
exports.__PRIVATE__wrapStartTransactionWithProfiling = __PRIVATE__wrapStartTransactionWithProfiling;
/**
 * Patches startTransaction and stopTransaction with profiling logic.
 * This is used by the SDK's that do not support event hooks.
 * @private
 */
function _addProfilingExtensionMethods() {
    const carrier = (0, core_1.getMainCarrier)();
    if (!carrier.__SENTRY__) {
        if ((0, env_1.isDebugBuild)()) {
            utils_1.logger.log("[Profiling] Can't find main carrier, profiling won't work.");
        }
        return;
    }
    carrier.__SENTRY__.extensions = carrier.__SENTRY__.extensions || {};
    if (!carrier.__SENTRY__.extensions['startTransaction']) {
        if ((0, env_1.isDebugBuild)()) {
            utils_1.logger.log('[Profiling] startTransaction does not exists, profiling will not work. Make sure you import @sentry/tracing package before @sentry/profiling-node as import order matters.');
        }
        return;
    }
    if ((0, env_1.isDebugBuild)()) {
        utils_1.logger.log('[Profiling] startTransaction exists, patching it with profiling functionality...');
    }
    carrier.__SENTRY__.extensions['startTransaction'] = __PRIVATE__wrapStartTransactionWithProfiling(
    // This is already patched by sentry/tracing, we are going to re-patch it...
    carrier.__SENTRY__.extensions['startTransaction']);
}
/**
 * This patches the global object and injects the Profiling extensions methods
 */
function addProfilingExtensionMethods() {
    _addProfilingExtensionMethods();
}
exports.addProfilingExtensionMethods = addProfilingExtensionMethods;
