"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ProfilingIntegration = void 0;
const utils_1 = require("@sentry/utils");
const env_1 = require("./env");
const hubextensions_1 = require("./hubextensions");
const utils_2 = require("./utils");
const utils_3 = require("./utils");
const MAX_PROFILE_QUEUE_LENGTH = 50;
const PROFILE_QUEUE = [];
const PROFILE_TIMEOUTS = {};
function addToProfileQueue(profile) {
    PROFILE_QUEUE.push(profile);
    // We only want to keep the last n profiles in the queue.
    if (PROFILE_QUEUE.length > MAX_PROFILE_QUEUE_LENGTH) {
        PROFILE_QUEUE.shift();
    }
}
// We need this integration in order to actually send data to Sentry. We hook into the event processor
// and inspect each event to see if it is a transaction event and if that transaction event
// contains a profile on it's metadata. If that is the case, we create a profiling event envelope
// and delete the profile from the transaction metadata.
class ProfilingIntegration {
    constructor() {
        this.name = 'ProfilingIntegration';
        this.getCurrentHub = undefined;
    }
    setupOnce(addGlobalEventProcessor, getCurrentHub) {
        this.getCurrentHub = getCurrentHub;
        const client = this.getCurrentHub().getClient();
        if (client && typeof client.on === 'function') {
            client.on('startTransaction', (transaction) => {
                const profile_id = (0, hubextensions_1.maybeProfileTransaction)(client, transaction, undefined);
                if (profile_id) {
                    const options = client.getOptions();
                    // Not intended for external use, hence missing types, but we want to profile a couple of things at Sentry that
                    // currently exceed the default timeout set by the SDKs.
                    const maxProfileDurationMs = 
                    // @ts-expect-error maxProfileDurationMs is not intended for external use
                    (options._experiments && options._experiments.maxProfileDurationMs) || hubextensions_1.MAX_PROFILE_DURATION_MS;
                    // Enqueue a timeout to prevent profiles from running over max duration.
                    if (PROFILE_TIMEOUTS[profile_id]) {
                        global.clearTimeout(PROFILE_TIMEOUTS[profile_id]);
                        delete PROFILE_TIMEOUTS[profile_id];
                    }
                    PROFILE_TIMEOUTS[profile_id] = global.setTimeout(() => {
                        if ((0, env_1.isDebugBuild)()) {
                            utils_1.logger.log('[Profiling] max profile duration elapsed, stopping profiling for:', transaction.name);
                        }
                        const profile = (0, hubextensions_1.stopTransactionProfile)(transaction, profile_id);
                        if (profile) {
                            addToProfileQueue(profile);
                        }
                    }, maxProfileDurationMs);
                    transaction.setContext('profile', { profile_id });
                    // @ts-expect-error profile is not part of txn metadata
                    transaction.setMetadata({ profile_id: profile_id });
                }
            });
            client.on('finishTransaction', (transaction) => {
                // @ts-expect-error profile is not part of txn metadata)
                const profile_id = transaction && transaction.metadata && transaction.metadata.profile_id;
                if (profile_id) {
                    if (PROFILE_TIMEOUTS[profile_id]) {
                        global.clearTimeout(PROFILE_TIMEOUTS[profile_id]);
                        delete PROFILE_TIMEOUTS[profile_id];
                    }
                    const profile = (0, hubextensions_1.stopTransactionProfile)(transaction, profile_id);
                    if (profile) {
                        addToProfileQueue(profile);
                    }
                }
            });
            client.on('beforeEnvelope', (envelope) => {
                var _a, _b;
                // if not profiles are in queue, there is nothing to add to the envelope.
                if (!PROFILE_QUEUE.length) {
                    return;
                }
                const profiledTransactionEvents = (0, utils_3.findProfiledTransactionsFromEnvelope)(envelope);
                if (!profiledTransactionEvents.length) {
                    return;
                }
                const profilesToAddToEnvelope = [];
                for (let i = 0; i < profiledTransactionEvents.length; i++) {
                    const profiledTransaction = profiledTransactionEvents[i];
                    const profile_id = (_b = (_a = profiledTransaction === null || profiledTransaction === void 0 ? void 0 : profiledTransaction.contexts) === null || _a === void 0 ? void 0 : _a['profile']) === null || _b === void 0 ? void 0 : _b['profile_id'];
                    if (!profile_id) {
                        throw new TypeError('[Profiling] cannot find profile for a transaction without a profile context');
                    }
                    // We need to find both a profile and a transaction event for the same profile_id.
                    const profileIndex = PROFILE_QUEUE.findIndex((p) => p.profile_id === profile_id);
                    if (profileIndex === -1) {
                        if ((0, env_1.isDebugBuild)()) {
                            utils_1.logger.log(`[Profiling] Could not retrieve profile for transaction: ${profile_id}`);
                        }
                        continue;
                    }
                    const cpuProfile = PROFILE_QUEUE[profileIndex];
                    if (!cpuProfile) {
                        if ((0, env_1.isDebugBuild)()) {
                            utils_1.logger.log(`[Profiling] Could not retrieve profile for transaction: ${profile_id}`);
                        }
                        continue;
                    }
                    // Remove the profile from the queue.
                    PROFILE_QUEUE.splice(profileIndex, 1);
                    const profile = (0, utils_3.createProfilingEvent)(cpuProfile, profiledTransaction);
                    if (profile) {
                        profilesToAddToEnvelope.push(profile);
                    }
                }
                (0, utils_2.addProfilesToEnvelope)(envelope, profilesToAddToEnvelope);
            });
        }
        else {
            // Patch the carrier methods and add the event processor.
            (0, hubextensions_1.addProfilingExtensionMethods)();
            addGlobalEventProcessor(this.handleGlobalEvent.bind(this));
        }
    }
    handleGlobalEvent(event) {
        if (this.getCurrentHub === undefined) {
            return (0, utils_3.maybeRemoveProfileFromSdkMetadata)(event);
        }
        if ((0, utils_3.isProfiledTransactionEvent)(event)) {
            // Client, Dsn and Transport are all required to be able to send the profiling event to Sentry.
            // If either of them is not available, we remove the profile from the transaction event.
            // and forward it to the next event processor.
            const hub = this.getCurrentHub();
            const client = hub.getClient();
            if (!client) {
                if ((0, env_1.isDebugBuild)()) {
                    utils_1.logger.log('[Profiling] getClient did not return a Client, removing profile from event and forwarding to next event processors.');
                }
                return (0, utils_3.maybeRemoveProfileFromSdkMetadata)(event);
            }
            const dsn = client.getDsn();
            if (!dsn) {
                if ((0, env_1.isDebugBuild)()) {
                    utils_1.logger.log('[Profiling] getDsn did not return a Dsn, removing profile from event and forwarding to next event processors.');
                }
                return (0, utils_3.maybeRemoveProfileFromSdkMetadata)(event);
            }
            const transport = client.getTransport();
            if (!transport) {
                if ((0, env_1.isDebugBuild)()) {
                    utils_1.logger.log('[Profiling] getTransport did not return a Transport, removing profile from event and forwarding to next event processors.');
                }
                return (0, utils_3.maybeRemoveProfileFromSdkMetadata)(event);
            }
            // If all required components are available, we construct a profiling event envelope and send it to Sentry.
            if ((0, env_1.isDebugBuild)()) {
                utils_1.logger.log('[Profiling] Preparing envelope and sending a profiling event');
            }
            const envelope = (0, utils_3.createProfilingEventEnvelope)(event, dsn);
            if (envelope) {
                transport.send(envelope);
            }
        }
        // Ensure sdkProcessingMetadata["profile"] is removed from the event before forwarding it to the next event processor.
        return (0, utils_3.maybeRemoveProfileFromSdkMetadata)(event);
    }
}
exports.ProfilingIntegration = ProfilingIntegration;
